--- ========================================================= ---
---  Things not to forget
--- ========================================================= ---
--[[
rawget
GetNumShapeShiftForms()
GetShapeShiftInfo(i)
SetActionSelfCastKey("NONE")
]]--
--- ========================================================= ---
---  Setup Revision, Version, Date and add to AceLibrary
--- ========================================================= ---
local VERSION = tonumber(("$Revision: 43238 $"):match("%d+"))

local Sphere = Sphere
Sphere_Buttons = Sphere:NewModule("Buttons", "AceEvent-2.0", "AceHook-2.1")
local self = Sphere_Buttons
if Sphere.revision < VERSION then
	Sphere.version = "r" .. VERSION
	Sphere.revision = VERSION
	Sphere.date = ("$Date: 2007-07-08 18:00:57 -0400 (Sun, 08 Jul 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end
Sphere.desc = "Adds Buttons to the Sphere"

--- ========================================================= ---
---  Setup Locals
--- ========================================================= ---
local DogTag = AceLibrary("DogTag-1.0")
local G = AceLibrary('Gratuity-2.0')
local Deformat = AceLibrary("Deformat-2.0")

local L = Sphere:L("Sphere_Buttons")

local data = {
	maxSpellID = nil,
	buttonCount = 0,
	cursorType = nil,
	cursorInfo1 = nil,
	cursorInfo2 = nil,
	cursorHasItem = false,
}

local buttonTypeValidate = {
	none = L["Normal"],
	stance = L["Stance/Stealth"],
	harmful = L["Harmful/Helpful"],
	modifier = L["Modifier"],
}

local buttonTypeValidateDesc = {
	none = L["Selecting this will have your Button and buttons associated with it to respond normally.  Changing the Type will cause the Button and all buttons associated with it to be deleted."],
	stance = L["Selecting this will have your Button and buttons associated with it change depending on your stance/stealth.  Changing the Type will cause the Button and all buttons associated with it to be deleted."],
	harmful = L["Selecting this will have your Button and buttons associated with it change depending on if your target is friendly or hostile.  Changing the Type will cause the Button and all buttons associated with it to be deleted."],
	modifier = L["Selecting this will have your Button and buttons associated with it change depending on a modifier (shift, alt, ctrl).  Changing the Type will cause the Button and all buttons associated with it to be deleted."],
}

local buttonUnitValidate = {
	player = L["Player"],
	target = L["Target"],
	targettarget = L["Target of Target"],
	focus = L["Focus"],
	pet = L["Pet"],
	party1 = L["Party Member1"],
	party2 = L["Party Member2"],
	party3 = L["Party Member3"],
	party4 = L["Party Member4"],
	none = L["Normal"],
}

local buttonUnitValidateDesc = {}

local buttonShownValidate = {
	always = L["Always"],
	never = L["Never"],
	combat = L["Combat"],
	nocombat = L["Not in Combat"],
	unit = L["Unit"],
}

local buttonShownValidateDesc = {}

--- ========================================================= ---
---  On Intitalize / Setup Database
--- ========================================================= ---
function Sphere_Buttons:OnInitialize()
	self.db = Sphere:AcquireDBNamespace("Buttons")
	Sphere:RegisterDefaults("Buttons", "char", {
		settings = {
			scale = 0.75,
			offSet = 65,
		},
		colors = {
			oor = {0.75, 0.10, 0.10, 1},
			oom = {0.10, 0.25, 1, 1},
			unusable = {0.50, 0.50, 0.50, 1},
			itemequippedborder = {0.10, 1, 0.10, 1},
			buttonhighlightborder = {0, 0.25, 1, 1},
		},
		
		SphereBar1 = {
			menu = false,
			empty = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar2 = {
			menu = false,
			empty = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar3 = {
			menu = false,
			empty = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar4 = {
			menu = false,
			empty = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar5 = {
			menu = false,
			empty = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar6 = {
			menu = false,
			empty = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar7 = {
			menu = false,
			empty = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar8 = {
			menu = false,
			empty = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar9 = {
			menu = false,
			empty = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
		SphereBar10 = {
			menu = false,
			empty = false,
			type = "none",
			unit = "none",
			shown = "always",
			useunit = nil,
			buttons = {},
			statemaps = {},
		},
	})
	
	data.maxSpellID = self:FindMaxID()
	Sphere_Buttons:CreateBars()
end

--- ========================================================= ---
---  Setup Options
--- ========================================================= ---
Sphere_Buttons:RegisterSphereOption({
	name = L["Buttons"],
	desc = L["Change the Buttons options."],
	type = 'group',
	args = {
		appearance = {
			name = L["Appearance"],
			desc = L["Change how the Buttons Appearance looks in general."],
			type = 'group',
			args = {
				colors = {
					name = L["Colors"],
					desc = L["Change the Colors that the Buttons use."],
					type = 'group',
					pass = true,
					get = function(value)
						return unpack(self.db.char.colors[value])
					end,
					set = "SetColor",
					args = {
						oor = {
							name = L["Out of Range"],
							desc = L["Color for when Out of Range."],
							type = 'color',
							hasAlpha = false,
						},
						oom = {
							name = L["Out of Mana"],
							desc = L["Color for when Out of Mana."],
							type = 'color',
							hasAlpha = false,
						},
						unusable = {
							name = L["Unusable"],
							desc = L["Color for when Unusable."],
							type = 'color',
							hasAlpha = false,
						},
						itemequippedborder = {
							name = L["Item Equipped Border"],
							desc = L["Color for your Item Equipped Border."],
							type = 'color',
							hasAlpha = false,
						},
						buttonhighlightborder = {
							name = L["Button Highlight Border"],
							desc = L["Color for you mouseover a Button."],
							type = 'color',
							hasAlpha = false,
						},
					},
				},
				size = {
					name = L["Size"],
					desc = L["Set how large the Buttons are."],
					type = 'range',
					min = 0.50,
					max = 2,
					step = 0.01,
					bigStep = 0.05,
					get = function()
						return self.db.char.settings.scale
					end,
					set = "SetScale"
				},
				offSet = {
					name = L["offSet"],
					desc = L["Set how much of an offSet the Buttons are."],
					type = 'range',
					min = 50,
					max = 150,
					step = 1,
					bigStep = 5,
					get = function()
						return self.db.char.settings.offSet
					end,
					set = "SetOffSet"
				},
			},
		},
		button1 = {
			name = L["Button 01"],
			desc = L["Change Button 1 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar1.menu
					end,
					set = function() 
						self.db.char.SphereBar1.menu = not(self.db.char.SphereBar1.menu);
						self:SetButtonMenu("1")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar1.type
					end,
					set = function(name) 
						self:SetButtonType(name, "1")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar1.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "1")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar1.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "1")
					end,
				},
			},
		},
		button2 = {
			name = L["Button 02"],
			desc = L["Change Button 2 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar2.menu
					end,
					set = function() 
						self.db.char.SphereBar2.menu = not(self.db.char.SphereBar1.menu);
						self:SetButtonMenu("2")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar2.type
					end,
					set = function(name) 
						self:SetButtonType(name, "2")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar2.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "2")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar2.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "2")
					end,
				},
			},
		},
		button3 = {
			name = L["Button 03"],
			desc = L["Change Button 3 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar3.menu
					end,
					set = function() 
						self.db.char.SphereBar3.menu = not(self.db.char.SphereBar3.menu);
						self:SetButtonMenu("3")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar3.type
					end,
					set = function(name) 
						self:SetButtonType(name, "3")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar3.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "3")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar3.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "3")
					end,
				},
			},
		},
		button4 = {
			name = L["Button 04"],
			desc = L["Change Button 4 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar4.menu
					end,
					set = function() 
						self.db.char.SphereBar4.menu = not(self.db.char.SphereBar4.menu);
						self:SetButtonMenu("4")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar4.type
					end,
					set = function(name) 
						self:SetButtonType(name, "4")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar4.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "4")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar4.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "4")
					end,
				},
			},
		},
		button5 = {
			name = L["Button 05"],
			desc = L["Change Button 5 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar5.menu
					end,
					set = function() 
						self.db.char.SphereBar5.menu = not(self.db.char.SphereBar5.menu);
						self:SetButtonMenu("5")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar5.type
					end,
					set = function(name) 
						self:SetButtonType(name, "5")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar5.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "5")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar5.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "5")
					end,
				},
			},
		},
		button6 = {
			name = L["Button 06"],
			desc = L["Change Button 6 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar6.menu
					end,
					set = function() 
						self.db.char.SphereBar6.menu = not(self.db.char.SphereBar6.menu);
						self:SetButtonMenu("6")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar6.type
					end,
					set = function(name) 
						self:SetButtonType(name, "6")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar6.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "6")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar6.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "6")
					end,
				},
			},
		},
		button7 = {
			name = L["Button 07"],
			desc = L["Change Button 7 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar7.menu
					end,
					set = function() 
						self.db.char.SphereBar7.menu = not(self.db.char.SphereBar7.menu);
						self:SetButtonMenu("7")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar7.type
					end,
					set = function(name) 
						self:SetButtonType(name, "7")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar7.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "7")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar7.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "7")
					end,
				},
			},
		},
		button8 = {
			name = L["Button 08"],
			desc = L["Change Button 8 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar8.menu
					end,
					set = function() 
						self.db.char.SphereBar8.menu = not(self.db.char.SphereBar8.menu);
						self:SetButtonMenu("8")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar8.type
					end,
					set = function(name) 
						self:SetButtonType(name, "8")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar8.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "8")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar8.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "8")
					end,
				},
			},
		},
		button9 = {
			name = L["Button 09"],
			desc = L["Change Button 9 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar9.menu
					end,
					set = function() 
						self.db.char.SphereBar9.menu = not(self.db.char.SphereBar9.menu);
						self:SetButtonMenu("9")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar9.type
					end,
					set = function(name) 
						self:SetButtonType(name, "9")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar9.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "9")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar9.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "9")
					end,
				},
			},
		},
		button10 = {
			name = L["Button 10"],
			desc = L["Change Button 10 options."],
			type = 'group',
			args = {
				menu = {
					name = L["Menu"],
					desc = L["Turns the Button into a menu."],
					type = 'toggle',
					get = function()
						return self.db.char.SphereBar10.menu
					end,
					set = function() 
						self.db.char.SphereBar10.menu = not(self.db.char.SphereBar10.menu);
						self:SetButtonMenu("10")
					end,
				},
				type = {
					name = L["Type"],
					desc = L["Choose the Type of the Button."],
					type = 'text',
					validate = buttonTypeValidate,
					validateDesc = buttonTypeValidateDesc,
					get = function()
						return self.db.char.SphereBar10.type
					end,
					set = function(name) 
						self:SetButtonType(name, "10")
					end,
				},
				unit = {
					name = L["Button Unit"],
					desc = L["Choose which Unit you want the Button to react to."],
					type = 'text',
					validate = buttonUnitValidate,
					validateDesc = buttonUnitValidateDesc,
					get = function()
						return self.db.char.SphereBar10.unit
					end,
					set = function(name) 
						self:SetButtonUnit(name, "10")
					end,
				},
				shown = {
					name = L["Button Shown"],
					desc = L["Choose which Event you want the Button to be shown."],
					type = 'text',
					validate = buttonShownValidate,
					validateDesc = buttonShownValidateDesc,
					get = function()
						return self.db.char.SphereBar10.shown
					end,
					set = function(name) 
						self:SetButtonShown(name, "10")
					end,
				},
			},
		},

	}

})

--- ========================================================= ---
---  On Enable
--- ========================================================= ---
function Sphere_Buttons:OnEnable()
	data.maxSpellID = self:FindMaxID()
	
	self:ScheduleRepeatingEvent(self.UpdateButton, TOOLTIP_UPDATE_TIME, self)
	self:ScheduleRepeatingEvent(self.FlashButton, ATTACK_BUTTON_FLASH_TIME, self)
	
	self:ScheduleEvent("Sphere_Buttons_PostEnable", self.OnPostEnable, 1, self)
end

--- ========================================================= ---
---  On Disable
--- ========================================================= ---
function Sphere_Buttons:OnDisable()
	self:UnregisterAllEvents();

end

--- ========================================================= ---
---  On Post Enable
--- ========================================================= ---
function Sphere_Buttons:OnPostEnable()
	for index=1, data.buttonCount do
		Sphere_Buttons:SetIcon(getglobal("SphereButton"..index))
	end
end

--- ========================================================= ---
--- Find Max Spell ID
--- ========================================================= ---
function Sphere_Buttons:FindMaxID()
	local id = 10
	while GetSpellTexture(id, BOOKTYPE_SPELL) do
		id = id + 1
	end
	return id
end

--- ========================================================= ---
--- Spell Name TO Spell ID
--- ========================================================= ---
function Sphere_Buttons:GetSpellInfo(spellName)
	for i = data.maxSpellID, 1, -1 do
		if (GetSpellName(i, BOOKTYPE_SPELL)) == spellName then
			return i
		end
	end
end

--- ========================================================= ---
--- Find Item Location
--- ========================================================= ---
function Sphere_Buttons:FindItemLocation(itemName)
	local b, s, i = nil;
	local iBag, iSlot = nil;
	local iInv = nil;
	local itemLoc, itemA, itemB = nil;
	
	for b=0,4 do
		for s=1,GetContainerNumSlots(b) do
			if GetContainerItemLink(b,s) then
				if string.find(GetContainerItemLink(b,s),itemName) then
					iBag = b
					iSlot = s
				end
			end
		end
	end
	
	if ((iBag == nil) and (iSlot == nil)) then
		for i=0,23 do
			if GetInventoryItemLink("player",i) then
				if string.find(GetInventoryItemLink("player",i),itemName) then
					iInv = i
				end
			end
		end
	end
	
	if ((iBag) and (iSlot)) then
		itemLoc = "Container"
		itemA = iBag
		itemB = iSlot
	elseif (iInv) then
		itemLoc = "Inventory"
		itemA = iInv
	end

	return itemLoc, itemA, itemB
end

--- ========================================================= ---
--- Skin Button
--- ========================================================= ---
function Sphere_Buttons:SkinButton(btn)
	
	local icon = getglobal(btn.."Icon")
	icon:ClearAllPoints();
	icon:SetWidth(28)
	icon:SetHeight(28)
	icon:SetPoint("CENTER", 0, 0);
	
	getglobal(btn):SetNormalTexture("")
	local background = getglobal(btn):GetNormalTexture();
	background:ClearAllPoints();
	background:SetWidth(48);
	background:SetHeight(48);
	background:SetPoint("CENTER", 0, 0);
		
	getglobal(btn):SetHighlightTexture("Interface\\Addons\\Sphere\\art\\button")
	local highlight = getglobal(btn):GetHighlightTexture();
	highlight:ClearAllPoints();
	highlight:SetWidth(48);
	highlight:SetHeight(48);
	highlight:SetPoint("CENTER", 0, 0);
	local r, g, b, a = unpack(self.db.char.colors.buttonhighlightborder)
	highlight:SetVertexColor(r, g, b, a)
	
	getglobal(btn):SetPushedTexture("Interface\\Addons\\Sphere\\art\\button")
	local pushed = getglobal(btn):GetPushedTexture();
	pushed:ClearAllPoints();
	pushed:SetWidth(48);
	pushed:SetHeight(48);
	pushed:SetPoint("CENTER", 0, 0);
	
	getglobal(btn):SetCheckedTexture("Interface\\Addons\\Sphere\\art\\button")
	local checked = getglobal(btn):GetCheckedTexture();
	checked:ClearAllPoints();
	checked:SetWidth(48);
	checked:SetHeight(48);
	checked:SetPoint("CENTER", 0, 0);
	
	getglobal(btn):CreateTexture(btn.."Overlay", "OVERLAY")
	getglobal(btn.."Overlay"):SetTexture("Interface\\Addons\\Sphere\\art\\button")
	local overlay = getglobal(btn.."Overlay")
	overlay:ClearAllPoints();
	overlay:SetWidth(48)
	overlay:SetHeight(48)
	overlay:SetPoint("CENTER", 0, 0);
	
	local cooldown = getglobal(btn.."Cooldown")
	cooldown:ClearAllPoints();
	cooldown:SetWidth(27)
	cooldown:SetHeight(27)
	cooldown:SetPoint("CENTER", 0, 0);
	cooldown:SetFrameLevel(2)
	
	local flash = getglobal(btn.."Flash")
	flash:ClearAllPoints();
	flash:SetWidth(28)
	flash:SetHeight(28)
	flash:SetPoint("CENTER", 0, 0);
	flash:SetAlpha(0.60)
	
	local hotkey = getglobal(btn.."HotKey")
	hotkey:SetDrawLayer("Overlay")
	
	local count = getglobal(btn.."Count")
	count:SetDrawLayer("Overlay")

end

--- ========================================================= ---
---  Set Color
--- ========================================================= ---
function Sphere_Buttons:SetColor(given_kind, r, g, b, a)
	local kind = given_kind
	local colors = self.db.char.colors[kind]
	
	if r then
		colors[1] = r
		colors[2] = g
		colors[3] = b
		colors[4] = a
	else
		r, g, b, a = unpack(colors)
	end
	
	if kind == "buttonhighlightborder" then
		for index=1, data.buttonCount do
			highlight = getglobal("SphereButton"..index):GetHighlightTexture();
			local r, g, b, a = unpack(self.db.char.colors.buttonhighlightborder)
			highlight:SetVertexColor(r, g, b, a)
		end
	end
	
end

--- ========================================================= ---
---  Set Scale
--- ========================================================= ---
function Sphere_Buttons:SetScale(value)
	if value then
		self.db.char.settings.scale = value
	else
		value = self.db.char.settings.scale
	end
	
	for index=1, data.buttonCount do
		getglobal("SphereBar"..index):SetScale(value)
	end
end

--- ========================================================= ---
---  Set offSet
--- ========================================================= ---
function Sphere_Buttons:SetOffSet(value)
	if value then
		self.db.char.settings.offSet = value
	else
		value = self.db.char.settings.offSet
	end
	
	local offSet = self.db.char.settings.offSet
	
	for index=1, data.buttonCount do
		radians = -math.rad(((index-1) * 36) - 90);
		xLoc = offSet * math.cos(radians)
		yLoc = offSet * math.sin(radians)
		getglobal("SphereBar"..index):SetPoint("CENTER", xLoc, yLoc);
	end
end

--- ========================================================= ---
---  Set Button Menu
--- ========================================================= ---
function Sphere_Buttons:SetButtonMenu(id)
	if id then
		if (self.db.char["SphereBar"..id].menu) then
			getglobal("SphereBar"..id):Show();
		end
	end
end

--- ========================================================= ---
---  Set Button Type
--- ========================================================= ---
function Sphere_Buttons:SetButtonType(value, id)
	if value then
		self.db.char["SphereBar"..id].type = value
	elseif id then
		value = self.db.char["SphereBar"..id].type
	end
	
	Sphere_Buttons:DeleteBar(id)

	if value == "harmful" then
		self.db.char["SphereBar"..id].useunit = 1
		getglobal("SphereBar"..id):SetAttribute("useunit", 1)
		
		if Sphere:HasModule('SubButtons') then
			getglobal("SphereSubBar"..id):SetAttribute("useunit", 1)
		end
	else
		self.db.char["SphereBar"..id].useunit = nil
		getglobal("SphereBar"..id):SetAttribute("useunit", nil)
		
		if Sphere:HasModule('SubButtons') then
			getglobal("SphereSubBar"..id):SetAttribute("useunit", nil)
		end
	end
	
	if value == "modifier" then
		self.db.char["SphereBar"..id].statemaps = {
			["modifier:shift"] = "s",
			["nomodifier"] = "n",
			["modifier:ctrl"] = "c",
			["modifier:alt"] = "a",
		}
		Sphere_Buttons:UpdateBar(getglobal("SphereBar"..id))
		
		if Sphere:HasModule('SubButtons') then
			Sphere_SubButtons:UpdateBar(getglobal("SphereSubBar"..id))
		end
	else
		self.db.char["SphereBar"..id].statemaps = {}
		Sphere_Buttons:UpdateBar(getglobal("SphereBar"..id))
		
		if Sphere:HasModule('SubButtons') then
			Sphere_SubButtons:UpdateBar(getglobal("SphereSubBar"..id))
		end
	end
	
end

--- ========================================================= ---
---  Set Button Unit
--- ========================================================= ---
function Sphere_Buttons:SetButtonUnit(value, id)
	if value then
		self.db.char["SphereBar"..id].unit = value
	elseif id then
		value = self.db.char["SphereBar"..id].unit
	end
	
	if value == "none" then
		value = ""
	end
	
	getglobal("SphereBar"..id):SetAttribute("unit", value)
	
	if Sphere:HasModule('SubButtons') then
		getglobal("SphereSubBar"..id):SetAttribute("unit", value)
	end
end

--- ========================================================= ---
---  Set Button Shown
--- ========================================================= ---
function Sphere_Buttons:SetButtonShown(value, id)
	if value then
		self.db.char["SphereBar"..id].shown = value
	elseif id then
		value = self.db.char["SphereBar"..id].shown
	end
	
	UnregisterUnitWatch(getglobal("SphereBar"..id))
	
	if Sphere:HasModule('SubButtons') then
		UnregisterUnitWatch(getglobal("SphereSubBar"..id))
	end
	
	if value == "unit" then
		RegisterUnitWatch(getglobal("SphereBar"..id))
		
		if Sphere:HasModule('SubButtons') then
			RegisterUnitWatch(getglobal("SphereSubBar"..id))
		end
	end
	
	getglobal("SphereBar"..id):SetAttribute("shown", value)
	
	if Sphere:HasModule('SubButtons') then
		getglobal("SphereSubBar"..id):SetAttribute("shown", value)
	end

	if not value or value == "always" or (value == "nocombat" and not InCombatLockdown()) then
		getglobal("SphereBar"..id):Show()
		
		if Sphere:HasModule('SubButtons') then
			getglobal("SphereSubBar"..id):Show()
		end
	else
		getglobal("SphereBar"..id):Hide()
		
		if Sphere:HasModule('SubButtons') then
			getglobal("SphereSubBar"..id):Hide()
		end
	end
end

--- ========================================================= ---
---  Create (10) Bars
--- ========================================================= ---
function Sphere_Buttons:CreateBars()

	local vals, xLoc, yLoc
	local offSet = self.db.char.settings.offSet

	for index=1, 10 do
		
		data.buttonCount = data.buttonCount + 1
		
		local name = "SphereBar"..index
		local anchor = "SphereMainFrame"
		
		local bar = CreateFrame("Frame", name, getglobal(anchor), "SecureStateHeaderTemplate")
		
		vals = self.db.char["SphereBar"..index]

		bar:SetID(index);
		bar:SetScale(self.db.char.settings.scale)
		bar:SetAlpha(1)
		
		radians = -math.rad(((index-1) * 36) - 90);
		xLoc = offSet * math.cos(radians)
		yLoc = offSet * math.sin(radians)
		bar:SetPoint("CENTER", xLoc, yLoc);
		
		if vals.unit == "none" then
			bar:SetAttribute("unit", "")
		else
			bar:SetAttribute("unit", vals.unit)
		end
		
		bar:SetAttribute("shown", vals.shown)
		bar:SetAttribute("useunit", vals.useunit)

		if vals.shown == "unit" then
			RegisterUnitWatch(bar)
		end

		if not vals.shown or vals.shown == "always" or (vals.shown == "nocombat" and not InCombatLockdown()) then
			bar:Show()
		else
			bar:Hide()
		end
		
		Sphere_Buttons:UpdateBar(bar)
		
	end
	
end

--- ========================================================= ---
---  Delete Bar
--- ========================================================= ---
function Sphere_Buttons:DeleteBar(id)
	local btn = getglobal("SphereButton"..id)
	
	local attribs = Sphere.db.namespaces.Buttons.char[btn:GetParent():GetName()].buttons[btn:GetID()]
	for k in pairs(attribs) do
		btn:SetAttribute(k, nil)
	end
	
	if Sphere:HasModule('SubButtons') then
		for i=1, 10 do
			local btn = getglobal("SphereButton"..id.."Sub"..i)
			
			local attribs = Sphere.db.namespaces.SubButtons.char[btn:GetParent():GetName()].buttons[btn:GetID()]
			for k in pairs(attribs) do
				btn:SetAttribute(k, nil)
			end
		end
	end

end

--- ========================================================= ---
--- Update Bar
--- ========================================================= ---
function Sphere_Buttons:UpdateBar(bar)
	local state
	local statelist, driver = {}, {}, {}

	local vals = self.db.char[bar:GetName()]
	if vals.statemaps then
		for k, v in pairs(vals.statemaps) do
			if string.find(k, "%-") then
				k = string.gsub(k, "%-", ":")
			end

			local add = true
			local list = v..":"..v
			table.insert(driver, "["..k.."] "..v)

			for _, l in ipairs(statelist) do
				if l == list then
					add = nil
					break
				end
			end

			if add then
				table.insert(statelist, v..":"..v)
			end
		end
	end

	driver = table.concat(driver, ";")
	state = SecureCmdOptionParse(driver)
	statelist = table.concat(statelist, ";")

	if #driver == 0 then
		state = "0"
	end

	RegisterStateDriver(bar, "state", driver)
	bar:SetAttribute("statemap-state", "$input")
	bar:SetAttribute("statebutton", statelist)
	bar:SetAttribute("state", state)
	bar:SetWidth(32)
	bar:SetHeight(32)

	local btn = Sphere_Buttons:CreateButton()
	btn:SetID(1)
	bar:SetAttribute("addchild", btn)

	vals.buttons[1] = vals.buttons[1] or {}
	for k, v in pairs(vals.buttons[1]) do
		btn:SetAttribute(k, v)
	end
	
	btn:SetAllPoints(bar);
	btn:SetFrameLevel(2)
	
	Sphere:AddToFrames(bar:GetName())
	
end

--- ========================================================= ---
--- Create Button
--- ========================================================= ---
function Sphere_Buttons:CreateButton()

	local btn = CreateFrame("CheckButton", "SphereButton"..data.buttonCount, UIParent, "SecureActionButtonTemplate, ActionButtonTemplate")
	btn:RegisterForClicks("AnyUp")
	btn:RegisterForDrag("LeftButton", "MiddleButton", "RightButton")
	btn:SetAttribute("checkselfcast", true)
	btn:SetAttribute("useparent-unit", true)
	btn:SetAttribute("useparent-statebutton", true)
	
	btn:SetScript("OnEnter", function(self) Sphere_Buttons:ButtonEnter(self) end)
	btn:SetScript("OnLeave", function(self) Sphere_Buttons:ButtonLeave(self) end)
	
	btn:SetScript("OnDragStart", function(self)
			if not InCombatLockdown() then
			
				local clickType = 1;
				if (arg1 == "RightButton") then
					clickType = 2;
				end
				if (arg1 == "MiddleButton") then
					clickType = 3;
				end
				Sphere_Buttons:ButtonDragStart(self, clickType)
			end
		end)
	
	btn:SetScript("OnReceiveDrag", function() end)
	
	btn:SetScript("PreClick", function(self)
			if GetCursorInfo() and not InCombatLockdown() then
				self:SetScript("OnClick", nil)
				
				local clickType = 1;
				if (arg1 == "RightButton") then
					clickType = 2;
				end
				if (arg1 == "MiddleButton") then
					clickType = 3;
				end
				
				local type, info1, info2
				
				if GetCursorInfo() then
					type, info1, info2 = GetCursorInfo()
				end
				Sphere_Buttons:ButtonReceiveDrag(self, clickType, type, info1, info2)
			end
		end)
		
	btn:SetScript("PostClick", function(self)
			self:SetChecked(nil)
			if not self:GetScript("OnClick") then
				self:GetParent():SetAttribute("addchild", self)
			end
		end)
	
	btn:SetScript("OnEvent", Sphere_Buttons.OnEvent);
	btn:RegisterEvent("PLAYER_ENTERING_WORLD");
	
	local cd = getglobal(btn:GetName().."Cooldown")
	cd:SetFrameLevel(cd:GetParent():GetFrameLevel() + 3)
	cd.text = cd:CreateFontString(nil, "OVERLAY")
	cd.text:SetPoint("CENTER")
	cd.text:SetFont("Fonts\\FRIZQT__.TTF", 22, "OUTLINE")
	
	local flash = getglobal(btn:GetName().."Flash")
	flash:ClearAllPoints()
	flash:SetAllPoints(cd)
	btn.flashtime = 0
	
	btn.style = "main"
		
	btn:SetScript("OnAttributeChanged", function(self, name, value)
			Sphere_Buttons:SaveAttrib(self, name, value)
		end)
		
	Sphere_Buttons:SkinButton(btn:GetName())
	
	return btn
end

--- ========================================================= ---
--- Button Enter
--- ========================================================= ---
function Sphere_Buttons:ButtonEnter(this)
	if GetCursorInfo() then
		local type, info1, info2 = GetCursorInfo()
		data.cursorHasItem = true
		data.cursorType =  type
		data.cursorInfo1 =  info1
		data.cursorInfo2 =  info2
	else
		data.cursorType =  nil
		data.cursorInfo1 =  nil
		data.cursorInfo2 =  nil
	end
end

--- ========================================================= ---
--- Button Leave
--- ========================================================= ---
function Sphere_Buttons:ButtonLeave(this)
	if not InCombatLockdown() and not Sphere.db.char.settings.locked then
		if (not(GetCursorInfo()) and (data.cursorHasItem)) then
			Sphere_Buttons:ButtonReceiveDrag(this, 2, data.cursorType, data.cursorInfo1, data.cursorInfo2)
			data.cursorHasItem = false
		end
	end

	GameTooltip:Hide() 
end

--- ========================================================= ---
--- Button Drag Start
--- ========================================================= ---
function Sphere_Buttons:ButtonDragStart(this, clickType)
	if not Sphere.db.char.settings.locked then
		local type, action = nil
	
		local bar = "SphereBar"..this:GetParent():GetID()

		if self.db.char[bar].type == "stance" then
		
		elseif self.db.char[bar].type == "modifiers" then
		
		elseif self.db.char[bar].type == "harmful" then

		else
			if self.db.char[bar].buttons[1]["*type"..clickType] then
				type = self.db.char[bar].buttons[1]["*type"..clickType]
				self.db.char[bar].buttons[1]["*type"..clickType] = nil
				if clickType == 1 then
					self.db.char[bar].buttons[1]["*type*"] = nil
					this:SetAttribute("*type*", nil)
				end
			end
			
			if type then
				action = self.db.char[bar].buttons[1]["*"..type..""..clickType]
				self.db.char[bar].buttons[1]["*"..type..""..clickType] = nil
				if clickType == 1 then
					self.db.char[bar].buttons[1]["*"..type.."*"] = nil
					this:SetAttribute("*"..action.."*", nil)
				end
			end
			
			if type == "spell" then
				PickupSpell(action)
			elseif type == "item" then
				local itemLoc, itemA, itemB = Sphere_Buttons:FindItemLocation(action)
				if itemLoc == "Container" then
					PickupContainerItem(itemA, itemB)
				elseif itemLoc == "Inventory" then
					PickupInventoryItem(itemA)
				end
			elseif type == "macro" then
				PickupMacro(action)
			end
		end
	end
end

--- ========================================================= ---
--- Button Drag Stop
--- ========================================================= ---
function Sphere_Buttons:ButtonReceiveDrag(this, clickType, type, info1, info2)
	if not Sphere.db.char.settings.locked then
		local state = this:GetAttribute("state-parent")
		if state == "0" then
			state = nil
		end
		
		if state then
			state = "-"..state
		else
			state = "*"
		end
		
		local unit = SecureButton_GetModifiedUnit(this, (state or "LeftButton"))
		local useunit = this:GetParent():GetAttribute("useunit")
		if useunit and unit and unit ~= "none" and UnitExists(unit) then
			if UnitCanAttack("player", unit) then
				this:SetAttribute("*harmbutton"..state, "nuke"..state)
				state = "-nuke"..state
			elseif UnitCanAssist("player", unit) then
				this:SetAttribute("*helpbutton"..state, "heal"..state)
				state = "-heal"..state
			end
		end
	
		if clickType == 1 then
			this:SetAttribute("*type*", type)
			if type == "spell" then
				local spell, rank = GetSpellName(info1, info2)
				this:SetAttribute("*spell"..state, spell)
				this:SetAttribute("*item"..state, nil)
				this:SetAttribute("*itemid"..state, nil)
				this:SetAttribute("*macro"..state, nil)
			elseif type == "item" then
				local item = GetItemInfo(info1)
				this:SetAttribute("*spell"..state, nil)
				this:SetAttribute("*item"..state, item)
				this:SetAttribute("*itemid"..state, info1)
				this:SetAttribute("*macro"..state, nil)
			elseif type == "macro" then
				this:SetAttribute("*spell"..state, nil)
				this:SetAttribute("*item"..state, nil)
				this:SetAttribute("*itemid"..state, nil)
				this:SetAttribute("*macro"..state, info1)
			end
		end
		
		local value = this:GetAttribute("value-parent")
		if value == "0" then
			value = nil
		end
		
		if value then
			value = "-"..value
		else
			value = clickType
		end
		
		local unit = SecureButton_GetModifiedUnit(this, (value or "LeftButton"))
		local useunit = this:GetParent():GetAttribute("useunit")
		if useunit and unit and unit ~= "none" and UnitExists(unit) then
			if UnitCanAttack("player", unit) then
				this:SetAttribute("*harmbutton"..value, "nuke"..value)
				value = "-nuke"..value
			elseif UnitCanAssist("player", unit) then
				this:SetAttribute("*helpbutton"..value, "heal"..value)
				value = "-heal"..value
			end
		end

		this:SetAttribute("*type"..value, type)
		if type == "spell" then
			local spell, rank = GetSpellName(info1, info2)
			this:SetAttribute("*spell"..value, spell)
			this:SetAttribute("*item"..value, nil)
			this:SetAttribute("*itemid"..value, nil)
			this:SetAttribute("*macro"..value, nil)
		elseif type == "item" then
			local item = GetItemInfo(info1)
			this:SetAttribute("*spell"..value, nil)
			this:SetAttribute("*item"..value, item)
			this:SetAttribute("*itemid"..value, info1)
			this:SetAttribute("*macro"..value, nil)
		elseif type == "macro" then
			this:SetAttribute("*spell"..value, nil)
			this:SetAttribute("*item"..value, nil)
			this:SetAttribute("*itemid"..value, nil)
			this:SetAttribute("*macro"..value, info1)
		end
		ClearCursor()
		
		data.cursorHasItem = false
		data.cursoractionType =  nil
		data.cursorInfo1 =  nil
		data.cursorInfo2 =  nil
	end
end

--- ========================================================= ---
--- Update Button
--- ========================================================= ---
function Sphere_Buttons:UpdateButton()
	for index=1, data.buttonCount do
		local btn = getglobal("SphereButton"..index)
		local r, g, b, a = 1, 1, 1, 1
		local unit = SecureButton_GetUnit(btn)
		if btn.type == "spell" then
			local usable, nomana = IsUsableSpell(btn.spell) -- need check here
			if nomana then
				r, g, b, a = unpack(self.db.char.colors.oom)
			elseif not usable then
				r, g, b, a = unpack(self.db.char.colors.unusable)
			elseif SpellHasRange(btn.spell) and IsSpellInRange(btn.spell, unit) == 0 then
				r, g, b, a = unpack(self.db.char.colors.oor)
			end
		elseif btn.type == "item" then
			if not IsUsableItem(btn.spell) then
				r, g, b, a = unpack(self.db.char.colors.unusable)
			elseif ItemHasRange(btn.spell) and IsItemInRange(btn.spell, unit) == 0 then
				r, g, b, a = unpack(self.db.char.colors.oor)
			end
		elseif btn.type == "macro" and GetMacroBody(btn.spell) then
			local action = GetActionFromMacroText(GetMacroBody(btn.spell))
			if action then
				if GetItemInfo(action) then
					if not IsUsableItem(action) then
						r, g, b, a = unpack(self.db.char.colors.unusable)
					elseif ItemHasRange(action) and IsItemInRange(action, unit) == 0 then
						r, g, b, a = unpack(self.db.char.colors.oor)
					end
				else
					local usable, nomana = IsUsableSpell(action)
					if nomana then
						r, g, b, a = unpack(self.db.char.colors.oom)
					elseif not usable then
						r, g, b, a = unpack(self.db.char.colors.unusable)
					elseif SpellHasRange(action) and IsSpellInRange(action, unit) == 0 then
						r, g, b, a = unpack(self.db.char.colors.oor)
					end
				end
			end
			Sphere_Buttons:SetIcon(btn)
		end
		getglobal(btn:GetName().."Icon"):SetVertexColor(r, g, b, a)
	end
end

--- ========================================================= ---
--- Set Icon
--- ========================================================= ---
function Sphere_Buttons:SetIcon(btn)
	local value = btn:GetAttribute("state-parent")
	if value then
		value = Sphere_Buttons:ConvertButton(btn, value)
		local texture, spell
		local text = getglobal(btn:GetName().."Name")
		local new = SecureButton_GetModifiedAttribute(btn, "type", value)
		btn.type = new
		btn.attack = nil
		text:SetText(nil)
		if new then
			spell = SecureButton_GetModifiedAttribute(btn, new, value)
			if spell then
				if new == "spell" then
					texture = GetSpellTexture(spell)
					if IsAttackSpell(spell) or IsAutoRepeatSpell(spell) then
						btn.attack = true
					end
				elseif new == "item" then
					local itemid = SecureButton_GetModifiedAttribute(btn, "itemid", value)
					if itemid then
						texture = select(10, GetItemInfo(itemid))
					end
				elseif new == "macro" then
					spell, texture = GetMacroInfo(spell)
				end
			end
			btn.spell = spell
		end
		
		local icon = getglobal(btn:GetName().."Icon")
		if icon:GetTexture() ~= texture then
			icon:SetTexture(texture)
			if texture then
				btn:SetAlpha(1)
				UIFrameFadeIn(icon, 0.50)
			end
			Sphere_Buttons:UpdateItem(btn)
			Sphere_Buttons:UpdateCooldown(btn)
		end
			
		if not texture and not btn.grid then
			btn:SetAlpha(0)
		end
	end
end

--- ========================================================= ---
--- Update Item
--- ========================================================= ---
function Sphere_Buttons:UpdateItem(btn)
	if btn.spell then
		local count = getglobal(btn:GetName().."Count")
		local border = getglobal(btn:GetName().."Border")

		if btn.type == "spell" then
			local spellid = Sphere_Buttons:GetSpellInfo(btn.spell)
			G:SetSpell(spellid, "spell")
			
			local reg
			local reagent = G:Find(SPELL_REAGENTS, 2, 99, false, true, false) and true
			
			if reagent then
				for i = 1, G:NumLines() do
					if (G:Find(SPELL_REAGENTS, i, i, false, true, false) and true) then
						reg = G:GetLine(i)
						reg= string.gsub(reg, SPELL_REAGENTS, "")
						if string.find(reg, "^|c") then
							_, _, reg = string.find(reg, "|c%x+(.*)|r")
						end
						reg = string.gsub(reg, "%(%)", "")
					end
				end
			end

			if reg then
				local num = GetItemCount(reg)
				if num > 1 then
					count:SetText(num)
				else
					count:SetText(nil)
				end
			else
				count:SetText(nil)
			end
			border:Hide()
			
		elseif btn.type == "item" then
			local num = GetItemCount(btn.spell)
			if num > 1 then
				count:SetText(num)
			else
				count:SetText(nil)
			end
		end
	end
end

--- ========================================================= ---
--- Convert Button
--- ========================================================= ---
function Sphere_Buttons:ConvertButton(btn, value)
	local unit = SecureButton_GetAttribute(btn, "unit")
	if unit and unit ~= "none" and UnitExists(unit) then
		if UnitCanAttack("player", unit) then
			value = SecureButton_GetModifiedAttribute(btn, "harmbutton", value) or value
		elseif UnitCanAssist("player", unit) then
			value = SecureButton_GetModifiedAttribute(btn, "helpbutton", value) or value
		end
	end
	return value
end

--- ========================================================= ---
--- Save Attributes
--- ========================================================= ---
function Sphere_Buttons:SaveAttrib(self, name, value)
	if string.find(name, "^%*") or string.find(name, "^shift") or string.find(name, "^ctrl") or string.find(name, "^alt") then
		Sphere.db.namespaces.Buttons.char[self:GetParent():GetName()].buttons[self:GetID()][name] = value
		name = "state-parent"
	end

	if name == "state-parent" then
		Sphere_Buttons:SetIcon(self)
	end
end

--- ========================================================= ---
--- Flash Button
--- ========================================================= ---
function Sphere_Buttons:FlashButton()
	for index=1, data.buttonCount do
		local btn = getglobal("SphereButton"..index)
		local flash = getglobal(btn:GetName().."Flash")
		if btn.flash and btn.attack then
			if flash:IsVisible() then
				flash:Hide()
			else
				flash:Show()
			end
		else
			flash:Hide()
		end
	end
end

--- ========================================================= ---
--- Update Cooldown
--- ========================================================= ---
function Sphere_Buttons:UpdateCooldown(btn)
	local start, duration, enable = 0, 0, 0
	local cooldown = getglobal(btn:GetName().."Cooldown")
	if btn.spell then
		if btn.type == "spell" then
			start, duration, enable = GetSpellCooldown(btn.spell)
		elseif btn.type == "item" then
			start, duration, enable = GetItemCooldown(btn.spell)
		elseif btn.type == "macro" then
			local action = GetActionFromMacroText(GetMacroBody(btn.spell))
			if action then
				if GetItemInfo(action) then
					start, duration, enable = GetItemCooldown(action)
				else
					start, duration, enable = GetSpellCooldown(action)
				end
			end
		end
	end
	if start and duration then
		CooldownFrame_SetTimer(cooldown, start, duration, enable)
	end
end

--- ========================================================= ---
--- Button OnEvent
--- ========================================================= ---
function Sphere_Buttons:OnEvent()
	if (event == "PLAYER_ENTERING_WORLD") then
		Sphere_Buttons:RegisterButtonEvents(this);
		return;
	end
	
	if (event == "PLAYER_REGEN_ENABLED") then
		local show = this:GetParent():GetAttribute("shown")
		if show == "nocombat" then
			this:GetParent():Show()
		elseif show == "combat" then
			this:GetParent():Hide()
		end
	elseif (event == "PLAYER_REGEN_DISABLED") then
		local show = this:GetParent():GetAttribute("shown")
		if show == "nocombat" then
			this:GetParent():Hide()
		elseif show == "combat" then
			this:GetParent():Show()
		end
	elseif (event == "PLAYER_TARGET_CHANGED") then
		if SecureButton_GetUnit(this) == "target" then
			Sphere_Buttons:SetIcon(this)
		end
	elseif (event == "PLAYER_FOCUS_CHANGED") then
		if SecureButton_GetUnit(this) == "focus" then
			Sphere_Buttons:SetIcon(this)
		end
	elseif (event == "MODIFIER_STATE_CHANGED") then
		Sphere_Buttons:SetIcon(this)
	elseif (event == "ACTIONBAR_UPDATE_USABLE") then
		Sphere_Buttons:UpdateCooldown(this)
	elseif (event == "UPDATE_INVENTORY_ALERTS") then
		Sphere_Buttons:UpdateCooldown(this)
	elseif (event == "ACTIONBAR_UPDATE_COOLDOWN") then
		Sphere_Buttons:UpdateCooldown(this)
	elseif (event == "UPDATE_MACROS") then
		Sphere_Buttons:SetIcon(this)
	elseif (event == "BAG_UPDATE") then
		Sphere_Buttons:UpdateItem(this)
	elseif (event == "UNIT_SPELLCAST_SUCCEEDED") then
		if this.type == "macro" then
			this.macro = true
		end
	elseif (event == "ACTIONBAR_SHOWGRID") then
		if not Sphere.db.char.settings.locked then
			if this.style == "main" then
				if Sphere.db.namespaces.Buttons.char[this:GetParent():GetName()].menu then
					this.grid = true
					this:SetAlpha(1)
				else
					this.grid = true
					this:SetAlpha(1)
				end
			elseif this.style == "sub" then
				if Sphere.db.namespaces.Buttons.char[this:GetParent():GetParent():GetName()].menu then
					this.grid = true
					this:SetAlpha(1)
				else
					this.grid = nil
					this:SetAlpha(0)
				end
			end
		end
	elseif (event == "ACTIONBAR_HIDEGRID") then
			if not Sphere.db.char.settings.locked then
			if this.style == "main" then
				if Sphere.db.namespaces.Buttons.char[this:GetParent():GetName()].menu then
					this.grid = nil
					this:SetAlpha(1)
				else
					this.grid = nil
					if not getglobal(this:GetName().."Icon"):GetTexture() then
						this:SetAlpha(0)
					end
				end
			elseif this.style == "sub" then
				if Sphere.db.namespaces.Buttons.char[this:GetParent():GetParent():GetName()].menu then
					this.grid = nil
					if not getglobal(this:GetName().."Icon"):GetTexture() then
						this:SetAlpha(0)
					end
				else
					this.grid = nil
					this:SetAlpha(0)
				end
			end
		end
	elseif (event == "PLAYER_ENTER_COMBAT") then
		this.flash = true
	elseif (event == "START_AUTOREPEAT_SPELL") then
		this.flash = true
	elseif (event == "PLAYER_LEAVE_COMBAT") then
		this.flash = nil
	elseif (event == "STOP_AUTOREPEAT_SPELL") then
		this.flash = nil
	elseif (event == "UNIT_FACTION") then
		if SecureButton_GetUnit(this) == arg1 then
			Sphere_Buttons:SetIcon(this)
		end
	end
end

--- ========================================================= ---
--- Register Button Events
--- ========================================================= ---
function Sphere_Buttons:RegisterButtonEvents(this)
	this:RegisterEvent("PLAYER_REGEN_ENABLED");
	this:RegisterEvent("PLAYER_REGEN_DISABLED");
	this:RegisterEvent("PLAYER_TARGET_CHANGED");
	this:RegisterEvent("PLAYER_FOCUS_CHANGED");
	this:RegisterEvent("MODIFIER_STATE_CHANGED");
	this:RegisterEvent("ACTIONBAR_UPDATE_USABLE");
	this:RegisterEvent("UPDATE_INVENTORY_ALERTS");
	this:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN");
	this:RegisterEvent("UPDATE_MACROS");
	this:RegisterEvent("BAG_UPDATE");
	this:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
	this:RegisterEvent("ACTIONBAR_SHOWGRID");
	this:RegisterEvent("ACTIONBAR_HIDEGRID");
	this:RegisterEvent("PLAYER_ENTER_COMBAT");
	this:RegisterEvent("START_AUTOREPEAT_SPELL");
	this:RegisterEvent("PLAYER_LEAVE_COMBAT");
	this:RegisterEvent("STOP_AUTOREPEAT_SPELL");
	this:RegisterEvent("UNIT_FACTION");
end